#!/usr/local/bin/perl -w
#
#
#  Author: Preslav Nakov
#          UC Berkeley
#          nakov@cs.berkeley.edu
#  
#  Description: Module Dice coefficient
#
#  Last modified: 06/25/03
#
#

package calc_dice;

	use strict;
	use porter;

    BEGIN {
        use Exporter   ();
        our (@ISA, @EXPORT);

        @ISA         = qw(Exporter);
        @EXPORT      = qw(	&CalcDice_Classic 
							&CalcDice_Modified_Unigram 
							&CalcDice_Modified_Bigram 
							&CalcDice_Modified_Bigram_Phrases
						);
    }
    
	##################################
	###                            ###
    ###   LOCAL MODULE VARIABLES   ###
	###                            ###
	##################################

	############################
	###   MODULE PARAMETERS  ###
	############################
	use constant STOPLIST_FILE => 'stoplist.txt';
	my $WORD_DEFINITION        =  "[A-Za-z0-9]+";

	#####################################
	###   INTERNAL MODULE CONSTANTS   ###
	#####################################
	use constant MODIFIED_UNIGRAM_DICE        => 0;
	use constant MODIFIED_BIGRAM_DICE         => 1;
	use constant MODIFIED_BIGRAM_DICE_PHRASES => 2;

	my %stopList = LoadStopList(STOPLIST_FILE);

################################################
###
###  Function:     LoadStopList()
###  Export:       NO
###  Description:  Loads the stop-words file into the 
###  Parameters:
###      1: $stopListFile - Text file containing a list of stop-words
###
	sub LoadStopList
	{	my $stopListFile = shift;
		my %stopList;
		open(STOPLIST, $stopListFile) or die ("Failed to open $stopListFile for reading!");
		while (<STOPLIST>) {
			chomp;
			die "Incorrect assumption!" if (!/^([a-z]+)$/o);
			$stopList{$1} = 1;
		}
		close(STOPLIST);
		return %stopList;
	} ### LoadStopList()

################################################
###
###  Function:     GetWordList()
###  Export:       NO
###  Description:  Builds a hash of the words in $text and their frequencies (in fact we need exist & !exist only)
###  Parameters:
###      1: $text - Text to extract the words from. A word is defined by $WORD_DEFINITION (see above).
###      2: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print the word list. This is useful when we want to understand why this score.
###
	sub GetWordList
	{	my ($text, $explain) = @_;
		my %wordHash;
		while ($text =~ s/($WORD_DEFINITION)//o) {
			$wordHash{lc $1}++;
		}
		if ($explain) { printf "\nWORDLIST (%d):", scalar keys %wordHash; foreach (sort keys %wordHash) { print " $_"; } print " \n"; }
		return %wordHash;
	} ### GetWordList()


################################################
###
###  Function:     GetNoStopWordsPorterStemList()
###  Export:       NO
###  Description:  Builds a hash of the stems in $text and their frequencies
###  Parameters:
###      1: $text - Text to extract the stems from. 
###			A word is defined by $WORD_DEFINITION (see above). 
###         Stop-words are dropped. The others are stemmed using the Porter stemmer.
###      2: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print the word list. This is useful when we want to understand why this score.
###
	sub GetNoStopWordsPorterStemList
	{	my ($text, $explain) = @_;
		my ($cnt, %stemHash);
		for ($cnt = 0; $text =~ s/($WORD_DEFINITION)//o;) {
			my $stem = porter(lc $1);
			if (!(exists $stopList{$stem})) {
				$stemHash{$stem}++;
				$cnt++;
			}
		}
		if ($explain) { print "STEMLIST ($cnt):"; foreach (sort keys %stemHash) { print " $_($stemHash{$_})"; } print " \n"; }
		return ($cnt, %stemHash);
	} ### GetNoStopWordsPorterStemList()


################################################
###
###  Function:     GetNoStopWordsPorterStemListBigramsPhrases()
###  Export:       NO
###  Description:  Builds a hash of the stems in $text and their frequencies
###  Parameters:
###      1: $text - Text to extract the stem bigrams from.
###			A word is defined by $WORD_DEFINITION (see above). 
###         Stop-words are dropped. The others are stemmed using the Porter stemmer.
###         Then bigrams are built.
###      2: $usePhrases - when != 0, we build a bigram from each subsequent pair of stems.
###         When == 0, we build a bigram only from subsequent stems that contain NO stopwords in between.
###      3: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print the word list. This is useful when we want to understand why this score.
###
	sub GetNoStopWordsPorterStemListBigramsPhrases
	{	my ($text,$usePhrases,$explain) = @_;
		my ($cnt, %bigramHash, $prev);
		for ($prev = (), $cnt = 0; $text =~ s/($WORD_DEFINITION)//o;) {
			my $stem = porter(lc $1);
			if (exists $stopList{$stem}) {
				$prev = () if ($usePhrases);
			}
			else {
				if (defined $prev) {
					$bigramHash{"$prev\_$stem"}++ ;
					$cnt++;
				}
				$prev = $stem;
			}
		}
		if ($explain) { print "BIGRAMS ($cnt):"; foreach (sort keys %bigramHash) { print " $_($bigramHash{$_})"; } print " \n"; }
		return ($cnt, %bigramHash);
	} ### GetNoStopWordsPorterStemListBigramsPhrases()


################################################
###
###  Function:     CalcDice_Modified_Dice()
###  Export:       NO
###  Description:  Calculates the modified Dice coefficient: unigram, bigram and bigram phrases
###  Parameters:
###      1: $text1 - Text to be compared to $text2
###      2: $text2 - Text to be compared to $text1
###      3: $bigramType - determines which kind of modified Dice coefficient to calculate
###			Possible values are: MODIFIED_UNIGRAM_DICE, MODIFIED_BIGRAM_DICE and MODIFIED_BIGRAM_DICE_PHRASES
###      4: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print a formula showing how the score is calculated. 
###          In addition, we pass this parameter to GetNoStopWordsPorterStemList()
###			 and GetNoStopWordsPorterStemListBigramsPhrases()
###
	sub CalcDice_Modified_Dice
	{	my ($text1, $text2, $bigramType, $explain) = @_;
		my ($elemCnt1, %hash1, $elemCnt2, %hash2, $overlapCnt);
		
		### 1. Get the terms sets
		if (MODIFIED_BIGRAM_DICE_PHRASES == $bigramType) {
			($elemCnt1, %hash1) = GetNoStopWordsPorterStemListBigramsPhrases($text1,1,$explain);
			($elemCnt2, %hash2) = GetNoStopWordsPorterStemListBigramsPhrases($text2,1,$explain);
		}
		elsif (MODIFIED_BIGRAM_DICE == $bigramType) {
			($elemCnt1, %hash1) = GetNoStopWordsPorterStemListBigramsPhrases($text1,0,$explain);
			($elemCnt2, %hash2) = GetNoStopWordsPorterStemListBigramsPhrases($text2,0,$explain);
		}
		elsif (MODIFIED_UNIGRAM_DICE == $bigramType) {
			($elemCnt1, %hash1) = GetNoStopWordsPorterStemList($text1,$explain);
			($elemCnt2, %hash2) = GetNoStopWordsPorterStemList($text2,$explain);
		}
		else {
			die "Incorrect value for the parameter \$bigramType";
		}
		if (0 == ($elemCnt1 + $elemCnt2)) {
			printf "\nOverlap: 0\n" if ($explain);
			return 0.0;
		}

		### 2. Find the overlap
		$overlapCnt = 0;
		foreach (keys %hash1) {
			$overlapCnt += ($hash1{$_} <= $hash2{$_} ? $hash1{$_} : $hash2{$_}) if (exists $hash2{$_});
		}

		### 3. Calculate the Modified Unigram Dice coefficient
		printf "\nOverlap: $overlapCnt\nModified Unigram Dice = 2.0 * $overlapCnt / ($elemCnt1 + $elemCnt2)\n" if ($explain);
		return 2.0 * $overlapCnt / ($elemCnt1 + $elemCnt2);

	} ### CalcDice_Modified_Dice()


################################################
###
###  Function:     CalcDice_Classic()
###  Export:       YES
###  Description:  Calculates the Classic Dice coefficient
###  Parameters:
###      1: $text1 - Text to be compared to $text2
###      2: $text2 - Text to be compared to $text1
###      3: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print a formula showing how the score is calculated. 
###          In addition, we pass this parameter to GetWordList().
###
	sub CalcDice_Classic
	{	my ($text1, $text2, $explain) = @_;
		my (%hash1, %hash2, $overlapCnt);
		
		### 1. Get the word sets
		%hash1 = GetWordList($text1,$explain);
		%hash2 = GetWordList($text2,$explain);

		### 2. Find the overlap
		$overlapCnt = 0;
		foreach (keys %hash1) {
			$overlapCnt++ if (exists $hash2{$_});
		}

		### 3. Calculate the classic Dice coefficient
		printf "\nOverlap: $overlapCnt\nClassic Dice = 2.0 * $overlapCnt / (%d + %d)\n", scalar keys %hash1, scalar keys %hash2 if ($explain);
		return 2.0 * $overlapCnt / ((keys %hash1) + (keys %hash2));
	} ### CalcDice_Classic()


################################################
###
###  Function:     CalcDice_Modified_Unigram()
###  Export:       YES
###  Description:  Calculates the modified Dice coefficient UNIGRAM
###  Parameters:
###      1: $text1 - Text to be compared to $text2
###      2: $text2 - Text to be compared to $text1
###      3: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print a formula showing how the score is calculated. 
###
	sub CalcDice_Modified_Unigram
	{	my ($text1, $text2, $explain) = @_;
		return CalcDice_Modified_Dice($text1,$text2,MODIFIED_UNIGRAM_DICE,$explain);
	} ### CalcDice_Modified_Dice_Unigram()


################################################
###
###  Function:     CalcDice_Modified_Bigram()
###  Export:       YES
###  Description:  Calculates the modified Dice coefficient BIGRAM
###  Parameters:
###      1: $text1 - Text to be compared to $text2
###      2: $text2 - Text to be compared to $text1
###      3: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print a formula showing how the score is calculated. 
###
	sub CalcDice_Modified_Bigram
	{	my ($text1, $text2, $explain) = @_;
		return CalcDice_Modified_Dice($text1,$text2,MODIFIED_BIGRAM_DICE,$explain);
	} ### CalcDice_Modified_Dice_Bigram()


################################################
###
###  Function:     CalcDice_Modified_Bigram_Phrases()
###  Export:       YES
###  Description:  Calculates the modified Dice coefficient BIGRAM PHRASES
###  Parameters:
###      1: $text1 - Text to be compared to $text2
###      2: $text2 - Text to be compared to $text1
###      3: $explain - Should we print the list of words extracted?
###          If $explain != 0, we print a formula showing how the score is calculated. 
###
	sub CalcDice_Modified_Bigram_Phrases
	{	my ($text1, $text2, $explain) = @_;
		return CalcDice_Modified_Dice($text1,$text2,MODIFIED_BIGRAM_DICE_PHRASES,$explain);
	} ### CalcDice_Modified_Dice_Bigram_Phrases()

    END {
       ################################
       ###   Free the memory used   ###
       ################################
       %stopList = ();
    }

1;  # return a true value from the file  
